
#Risk and Hedonics in Empirical Agent-based land market model
# This model simulates the impact of flood risk in housing markets in flood-prone cities
#    Copyright (C) 2019 Koen de Koning and Tatiana Filatova

#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.

library(sp, lib)#.loc=tempdir())
library(rgdal)#, lib.loc=tempdir())
library(maptools)#, lib.loc=tempdir())
library(rgeos)#, lib.loc=tempdir())
library(GISTools)#, lib.loc=tempdir())
library(plotrix)#, lib.loc=tempdir())
library(foreign)#, lib.loc=tempdir())
library(gstat)#, lib.loc=tempdir())


#------------
# Settings  - These are the variables that can be changed for the purpose of running experiments. Changing these will not change the structure of the model.
#------------

Budgets <- "Empirical"                  #Choose Empirical or Match-Ini-Price, 
                                        # Empirical budgets are based on US wide income distributions and
                                        # housing budgets per income group (sources: Quigley and Raphael 2004, US sensus bureau and Statista.com)
                                        # Match-Ini-Price matches the buyers' housing budgets with the initial property values of the case study

RandomizedParcels <- TRUE               #Choose TRUE or FALSE
                                        # Defines whether the parcels' attributes (age, location, number of bedrooms etc.) match with empirical data or whether they are randomised
                                        # Choose TRUE when you want to prevent multicolinearity of the property attributes in the hedonic price estimation

fraction_on_sale <- 0.1                 #choose value between 0 and 1
                                        # Defines the fraction of properties that go on sale each time step of half a year. 
                                        # 10% means that people spend 0.5/0.10 = 5 years on average in a home before they sell again.

sd_fraction_on_sale <- 0.01             #choose value between 0 and 0.3
                                        # Here you can add semi-annual variability to the fraction of propeties that go on sale.
                                        # This represents the variability that you will typically see when external economic conditions could be favourable or unfavourable for selling/moving.

NofBuyers <- "Equal to sellers"         #choose Equal to selles, Empirical fixed or Empirical Random
                                        # Here you can specify how the number of buyers is selected at each time step. It can be fixed or random based on empirical data, or equal to sellers.

newBuyerCoef <- 0.7                     #best to leave it like this
                                        # specify the percentage of buyers that are entirely new to the area versus the buyers that already lived in the area. This is based on empirical evidence (survey data).

yM <- 30                                #best to leave it like this
                                        # specify the years payback time on a mortgage. In the U.S. this is typically 30 years.

aDelta <- 1                             #best to leave it like this
                                        # specify the time steps of rent that the sellers take into account when negotiating over prices. Would it be better to accept a deal now or wait for a better deal and accept paying N periods of rent?

alpha <- 0.10                           #choose value between 0 and 0.30
                                        # specify the relative price adjustment for properties that are either difficult to sell or that are easy to sell. The model takes into account that some houses in certain neighbourhoods or with certain attributes are demanded more than others.

SIMlength <- 120                        #choose value, ideally between 120 and 200
                                        # defines the number of time steps in the simulations
                                        # the first 100 time steps are needed for stabilising the model. The first flood will occur at t=102, and the second at t=110

NREPS <- 30                             #choose number of simulations (or number of repetitions of the same settings in case of running an experiment)

nN <- 5                                 #choose the number of houses a buyer considers in his search
                                        # After thorough testing it is advised to leave it at 5. This number represents the number of properties that are compared by each buyer.
                                        # i.e. a buyer does not compare all properties in his budget range but randomly selects a few that he will compare

UPDATEGRAPH <- FALSE                    #choose TRUE or FALSE, TRUE may slow down simulations
                                        # there are no spectacular graphics of the model dynamics to show at this time

fAVOIDRISK <- 0.2                       #which fraction of buyers completely avoid flood-zone properties (check survey results for the emperical estimation)
                                        # in the current version of the model this is agent-specific a depends on a number of attributes that are specified in survey data.
                                        # hence, changing this variable will have no effect on the simulations. Deleting this variable does require changes in the rest of the code.

Experiment <- TRUE                      #choose TRUE or FALSE. Decide whether you want to run an experiment. Check experimental settings below.

StudyArea <- "Greenville"               #Choose "Greenville" or "Beaufort"
                                        # Greenville is an inland area with a river that is subject to flooding. Beaufort is a coastal area that is subject to storm surge.
                                        # The amenities associated with water (beach, ocean view, private boat pier) are present only in Beaufort. Beaufort also has more properties in the flood zone.

Preferences <- "Heterogeneous"          #Choose whether buyers are "Homogeneous" or "Heterogeneous" in their preferences for housing characteristics
                                        # Besides heterogeneous budgets, buyers could also have different preferences for certain neighbourhoods or housing attributes.

Insurance <- TRUE                       #choose TRUE or FALSE
                                        # Mandatory insurances impose extra costs on property owners in the flood zone, and it affects households' tendency to avoid the flood zone.

Expectations <- "Dynamic"               #choose "Static" or "Dynamic"
                                        # Determines whether the realtor's price expectations adapt to changing preferences, for example when demand for risky properties decreases. When set at static, it will not respond to market conditions.

Scenario <- "Control"                   #choose 'Control', 'Flood', 'Extreme flood', 'Flood and media' or 'Media'
                                        # The control scenario has no flooding. All experiments will be compared with the control scenario.
                                        # A random seed function will ensure that all experiments can be compared from the moment that scenarios start to diverge.
                                        # The flood scenario has a single flood at t=102
                                        # The extreme flood scenario has two floods: at t=102 and t=110
                                        # The media scenario adds a reptitive news update where agents are informed about flooding through media sources.
                                        # It is assumed that the media also plays a role in the year of the flood.

NeighbourR <- 2000                      #choose in which radius (in feet) the neighbours are selected
                                        # this radius determines the neighbourhood within which households exchange information about whether their house got affected by a flood.

dir <- "P:/BMS/Projects/KOEN/RHEA"            #specify the directory where the model R code is stored
DATAdir <- "P:/BMS/Projects/KOEN/RHEA/Data"   #specify the directory where the input data of the model is stored
outputDir <- paste0(dir, "/Output")           #specify the directory where the output of the model will be stored
setwd(dir)

#----------------------
# Experiment settings - This part of the code is convenient when you want to run a number of experiments (full factorial)
#----------------------
# This code ensures that the model runs with all possible combinations of input variables that you vary
# in the experiments.

EXPsettings <- list(Budgets,RandomizedParcels,fraction_on_sale,Scenario,NofBuyers,
                    newBuyerCoef,nN,fAVOIDRISK,NeighbourR)
Design <- "Full factorial"      #could be changed in the future if many interactions will be tested
vary <- function(VarName, min=NA, max=NA, levels=NA) {
  if (VarName == "Budgets") {
    EXPsettings[[1]] <- levels
    if(is.na(levels[1])){
      print("please define levels for variable 'Budgets'")
      break}
    return(EXPsettings)}
  if (VarName == "RandomizedParcels") {
    EXPsettings[[2]] <- c(TRUE,FALSE)
    return(EXPsettings)}
  if (VarName == "fraction_on_sale") {
    EXPsettings[[3]] <- min + 0:(levels-1) * (max-min)/(levels-1)
    return(EXPsettings)}
  if (VarName == "Scenario") {
    EXPsettings[[4]] <- levels
    if(is.na(levels[1])){
      print("please define levels for variable 'Scenario'")
      break}
    return(EXPsettings)}
  if (VarName == "NofBuyers") {
    EXPsettings[[5]] <- levels
    if(is.na(levels[1])){
      print("please define levels for variable 'NofBuyers'")
      break}
    return(EXPsettings)}
  if (VarName == "newBuyerCoef") {
    EXPsettings[[6]] <- min + 0:(levels-1) * (max-min)/(levels-1)
    return(EXPsettings)}
  if (VarName == "nN") {
    EXPsettings[[7]] <- min + 0:(levels-1) * (max-min)/(levels-1)
    return(EXPsettings)}
  if (VarName == "fAVOIDRISK") {
    EXPsettings[[8]] <- min + 0:(levels-1) * (max-min)/(levels-1)
    return(EXPsettings)}
  if (VarName == "NeighbourR") {
    EXPsettings[[9]] <- min + 0:(levels-1) * (max-min)/(levels-1)
    return(EXPsettings)}
}

# choose the variables that you would like to vary in the experiment:
# a few other examples are listed below. You need to run each line separately.
EXPsettings <- vary("Scenario", levels=c('Control', 'Flood', 'Extreme flood'))
#EXPsettings <- vary("NeighbourR", min=1000, max=3000,levels=3)
#EXPsettings <- vary("fraction_on_sale", min=.02, max=.20,levels=5)
#EXPsettings <- vary("fAVOIDRISK", min=0, max=0.8,levels=7)


#------------------
# Setup Landscape
#------------------
setwd(DATAdir)

# landscape of the study area is imported in R
if (StudyArea == "Greenville") {
  CityMap <- readOGR('.', 'Greenville')
  PreFloyd <- readOGR('.','Greenville_Random3')
}
if (StudyArea == "Beaufort") {
  CityMap <- readOGR('.', 'Beaufort')
  CityMap <- CityMap[CityMap$LAND==1,]
  PreFloyd <- readOGR('.', 'Beaufort Random')
  #PreFloyd <- PreFloyd[apply(PreFloyd@data,1,FUN=function(x){!any(is.na(x))}),]
}

# list of household incomes are imported
income <- read.table("incomes.txt")$x

# list of preferences for certain housing attributes is imported. It is a file with weights for the various attributes.
BuyerPrefs <- read.table("Buyer preferences.csv", sep=";", header=TRUE)
if (Preferences=="Homogeneous") {
  BuyerPrefs <- round(apply(BuyerPrefs, 2, mean),1)
  names <- names(BuyerPrefs)                                 
  BuyerPrefs <- data.frame(matrix(BuyerPrefs, 1,5))
  names(BuyerPrefs) <- names}
if (Preferences=="Heterogeneous") {BuyerPrefs <- BuyerPrefs[-1,]}
TimeEst <- read.table("SimTimeEst.csv") # document with coefficients with which the simulation run time can be estimated

# survey data is imported, the data is used as empirical input for the households' initial settings, preferences and perceptions
SurveyDATA <- read.table("SurveyDATA.csv", sep=";", header=TRUE)

# add functions for the model's functionality
PFLOOD <- function(pflood) {
  YearsofResidence <- round(1/fraction_on_sale)
  NofOccurrence <- matrix(0:3, length(pflood), 4, byrow=TRUE)
  NoverK <- factorial(YearsofResidence)/(factorial(NofOccurrence)*factorial(YearsofResidence-NofOccurrence))
  pflood <- matrix(pflood, length(pflood), 4)
  Pflood <- pflood^NofOccurrence*(1-pflood)^(YearsofResidence-NofOccurrence)*NoverK
  return(Pflood)
}   # returns a matrix of flood probabilities related to N(0,1,2,3) occurrences and flood probability of a property
SIMcode <- function() {ALPHABET <- c("A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z")
alphabet <- c("a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z")
paste(ifelse(runif(1)<.5,sample(alphabet,1),sample(ALPHABET)),
      ifelse(runif(1)<.5,sample(alphabet,1),sample(ALPHABET)),
      ifelse(runif(1)<.5,sample(alphabet,1),sample(ALPHABET)),
      sample(0:9,1),ifelse(runif(1)<.5,sample(alphabet,1),sample(ALPHABET)),
      sample(0:9,1),sample(0:9,1),
      ifelse(runif(1)<.5,sample(alphabet,1),sample(ALPHABET)),
      ifelse(runif(1)<.5,sample(alphabet,1),sample(ALPHABET)),
      sep="")}    # generates a random code as a unique ID for a simulation
# Function 'Bayesian()' returns the bayesian probability of a specified dependent variable (e.g.: fear score), 
# given the independent variables. In case of fear score these should be: learing from
# news media, not learning about flood risk, and damage from flooding.
Bayesian <- function(SurveyDATA, DV, IV) {
  DATA <- SurveyDATA[IV]
  DATA$freq <- 1/nrow(DATA)
  P.ALLiv <- aggregate(freq~.,DATA,FUN=sum)
  SurveyDATA[DV] <- round(SurveyDATA[DV],1)
  DVDATA <- SurveyDATA[DV]
  DVDATA$freq <- 1/nrow(DVDATA)
  P.DV <- aggregate(freq~.,DVDATA,FUN=sum)
  
  pDV.IV <- list(NULL)
  for (K in 1:length(IV)) {
    DV.ivK <- SurveyDATA[c(IV[K],DV)]
    DV.ivK$freq <- 1/nrow(DV.ivK)
    pDV.ivK <- aggregate(freq~.,DV.ivK,FUN=sum)
    temp <- NULL
    for (i in unique(pDV.ivK[,IV[K]])) {
      freq <- rep(0,nrow(P.DV))
      rules <- P.DV[,DV] %in% pDV.ivK[DV][pDV.ivK[IV[K]]==i]
      freq[rules] <- pDV.ivK$freq[pDV.ivK[IV[K]]==i]
      temp <- cbind(temp,freq)
    }
    row.names(temp) <- P.DV[,DV]
    colnames(temp) <- unique(pDV.ivK[,IV[K]])
    pDV.ivK <- temp
    pDV.IV[[K]] <- pDV.ivK
  }
  
  P.ALL <- matrix(NA,nrow(P.DV),nrow(P.ALLiv))
  row.names(P.ALL) <- P.DV[,DV]
  for (i in 1:nrow(P.ALLiv)) {
    product <- 1
    for (K in 1:length(IV)) {
      temp <- pDV.IV[[K]][,as.character(P.ALLiv[i,IV[K]])]
      product <- product*temp
    }
    P.ALL[,i] <- product/P.DV$freq^(K-1)
  }
  sum(P.ALL)
  P.ALLiv$freq <- apply(P.ALL,2, FUN="sum")
  pDV.ALL <- apply(P.ALL,1,FUN=function(x){x/P.ALLiv$freq})
  pDV.ALL <- as.data.frame(pDV.ALL)
  P.ALLiv <- cbind(P.ALLiv,pDV.ALL)
  return(P.ALLiv)
}
# homeowners' perceived likelihood that their house will get damage from flooding:
LikelyDamage.Sel <- Bayesian(SurveyDATA[SurveyDATA$MarketStatus=="Seller",],
                             DV="LikelyDamage",IV=c("House.PFLOOD","Learning.Nghbrs","Exp.Damage2"))
# buyers' perceived likelihood that the house they are interested in buyer will get damage from flooding:
LikelyDamage.Buy <- Bayesian(SurveyDATA[SurveyDATA$MarketStatus=="Buyer",],
                             DV="LikelyDamage",IV=c("House.PFLOOD","Learning.NewsMedia","Exp.Damage2"))
# households' fear towards flood risk
FEARscore <- Bayesian(SurveyDATA,DV="FEARscore",IV=c("Learning.Not","Learning.NewsMedia","Exp.Damage2"))
# buyers' tendency to avoid buying a property that is located in the flood zone
BEHAVscore <- Bayesian(SurveyDATA,DV="BEHAVscore",IV=c("FEARscore","Flood.issue"))
# value reduction of flood zone properties immediately after the flood to simulate damages
RedAfDMG <- data.frame(mu=-0.137543,sigma=0.084689)
# additional value reduction for properties in the 100-year flood zone to simulate heavier damages and anticipated drop in demand for these properties.
RedAfFLD100 <- data.frame(mu=-0.160241,sigma=0.120906)

#-------------------
# Run Simulations 
#-------------------

setwd(dir)

# continue simulations from the last:
SIM <- 0
TEMP <- TRUE
while (TEMP) {
  
  SIM <- SIM + 1
  SEED <- (SIM-1)%/%nrow(SettingsMatrix)
  set.seed(SEED)
  Time <- Sys.time()
  SIM.ID <- SIMcode()
  
  index <- SIM%%dim(SettingsMatrix)[1]
  if(index==0){index <- dim(SettingsMatrix)[1]}
  Scenario <- as.character(SettingsMatrix$Scenario[index])
  
  SIM.ID <- paste0(substr(SIM.ID,1,4),"_",Scenario,"_",alpha)
  print(SIM.ID)
  
  TEMP <- file.exists(paste0(outputDir,"/SIM_ID_",SIM.ID,".csv"))
  print(paste("File exists =",TEMP))
  print(paste("Random seed =",SEED))
  
}
SIM1 <- SIM
remove(TEMP)
# prevents file over-writing because of the random seed function
# repeating the experiments with an empty output folder will produce exactly the same results. Any additional simulations will be added as a new file in the destination folder.

# when running experiments this function ensures that all the model settings can be updated in each repetition.
if(Experiment) {
  if(Design=="Full factorial") {
    TEMP <- 1
    for (VA in 1:length(EXPsettings)) {
      TEMP <- length(EXPsettings[[VA]]) * TEMP}
  }
  NSIM <- NREPS * TEMP
  
  step1 <- merge(EXPsettings[[1]],
                 EXPsettings[[2]])
  currentMATRIX <- step1
  for (VA in 3:length(EXPsettings)) {
    
    TEMP <- length(EXPsettings[[VA]])
    
    intermediate <- currentMATRIX
    INTERMEDIATE <- NULL
    for (merging in 1:TEMP) {
      INTERMEDIATE <- rbind(INTERMEDIATE,intermediate)
    }
    ADD <- merge(currentMATRIX[,1],EXPsettings[[VA]])[,2]
    
    currentMATRIX <- cbind(INTERMEDIATE, ADD)
    
  }
  names(currentMATRIX) <- c("Budgets","RandomizedParcels",'fraction_on_sale','Scenario','NofBuyers',
                            'newBuyerCoef','nN','fAVOIDRISK','NeighbourR')
  SettingsMatrix <- currentMATRIX
} else {NSIM <- NREPS}

# this entire section runs the simulations and writes the output files in the destination folder.
for (SIM in 1:NSIM+SIM1-1) {
  #-------------
  # Setup model - code for the initial setup of the model
  #-------------
  TICK <- 0
  
  set.seed((SIM-1)%/%nrow(SettingsMatrix))
  Time <- Sys.time()
  SIM.ID <- SIMcode()
  print("------------------------------------")
  print(paste("-- Stating simulation number", SIM, "--"))
  print("------------------------------------")
  print(Time)
  kH <- 2/yM
  kY <- 2
  
  # update settings if you are running an experiment
  if(Experiment) {
    index <- SIM%%dim(SettingsMatrix)[1]
    if(index==0){index <- dim(SettingsMatrix)[1]}
    Budgets <- SettingsMatrix$Budgets[index]
    RandomizedParcels <- SettingsMatrix$RandomizedParcels[index]
    fraction_on_sale <- SettingsMatrix$fraction_on_sale[index]
    Scenario <- as.character(SettingsMatrix$Scenario[index])
    NofBuyers <- SettingsMatrix$NofBuyers[index]
    newBuyerCoef <- SettingsMatrix$newBuyerCoef[index]
    nN <- SettingsMatrix$nN[index]
    fAVOIDRISK <- SettingsMatrix$fAVOIDRISK[index]
    NeighbourR <- SettingsMatrix$NeighbourR[index]
  }
  # value reduction of flood zone properties immediately after the flood to simulate damages
  DMGreduction <- rnorm(1,RedAfDMG$mu,RedAfDMG$sigma)
  # additional value reduction for properties in the 100-year flood zone to simulate heavier damages and anticipated drop in demand for these properties.
  EXPreduction <- rnorm(1,RedAfFLD100$mu,RedAfFLD100$sigma)
  
  # Tracking ID for the simulation runs. This can be changed by the user so that different experiments can be compared directly in the model output on the bases of the different input variables.
  SIM.ID <- paste0(substr(SIM.ID,1,4),"_",Scenario,"_",alpha)
  
  T0 <- Sys.time()
  
  # to save settings later
  if (file.exists(paste0(outputDir,"/Settings.csv"))) {
    print("READING FILES")
    Settings <- read.csv(paste0(outputDir,"/Settings.csv"))
    Settings$Time <- as.POSIXct(Settings$Time)
  } else {Settings <- NULL}
  Settings <- Settings[c("SIM.ID", "Time","Budgets","fraction_on_sale","Scenario","NofBuyers","newBuyerCoef",
                         "yM", "alpha", "SIMlength", "nN", "fAVOIDRISK","StudyArea","Preferences","Insurance","NeighbourR","DMGreduction","EXPreduction")]
  
  # make a data frame to store the settings of each model run
  SettingsSIMN <- data.frame(SIM.ID, Time,Budgets,fraction_on_sale,Scenario,NofBuyers,newBuyerCoef,
                             yM, alpha, SIMlength, nN, fAVOIDRISK,StudyArea,Preferences,Insurance,NeighbourR,DMGreduction,EXPreduction)
  Settings <- rbind(Settings, SettingsSIMN)
  write.table(Settings, paste0(outputDir,"/Settings.csv"), sep=",")
  Telapsed <- Sys.time()-T0
  # check how much time it takes to overwrite the 'Settings.csv' file. This important to avoid conflicts when running multiple experiments in parallel.
  print(Telapsed)
  
  # run the initial hedonic regression to estimate initial housing values.
  if (!is.null(PreFloyd$dflood)) {hedonics <- lm(LNPRICE~age+log(TOTALSQFT)+LNACRES+bedrm+dflood, PreFloyd)}
  if (!is.null(PreFloyd$dflood100)&
      !is.null(PreFloyd$dflood500)) {hedonics <- lm(LNPRICE~age+log(TOTALSQFT)+LNACRES+bedrm+dflood100+dflood500, PreFloyd)}
  PreFloyd$RESID <- resid(hedonics)
  PreFloyd$PRED <- predict(hedonics)
  HED.coef <- coef(hedonics)
  HED.coef.flooddummy <- HED.coef
  hedonics.nodummy <- lm(LNPRICE~age+log(TOTALSQFT)+LNACRES+bedrm, PreFloyd)
  HED.coef.nodummy <- coef(hedonics.nodummy)
  
  # the semivariogram that is created here is needed for spatial interpolation (kriging) of the residuals after hedonic regression
  vgmRESID <- variogram(RESID~1, PreFloyd)
  vgm.fit.RESID <- fit.variogram(vgmRESID, model = vgm(0, "Sph", 5000))
  #plot(vgmRESID, vgm.fit.RESID, main="(Semi)variogram residuals")
  
  plot(CityMap, col="orange", main=paste("Simulation number", SIM))
  plot(PreFloyd, pch=15, col="dark red", add=TRUE, cex=0.3)
  
  # create the parcels and attach their attributes
  Parcels <- PreFloyd
  Parcels <- Parcels[,c("ID","AGEnorm","SQFTnorm","ACRESnorm","BEDRMnorm","RESIDnorm")]
  if (!is.null(PreFloyd$dflood)) {Parcels$dflood <- PreFloyd$dflood}
  if (!is.null(PreFloyd$dflood100)&
      !is.null(PreFloyd$dflood500)) {Parcels$dflood100 <- PreFloyd$dflood100
      Parcels$dflood500 <- PreFloyd$dflood500}
  NParcels <- dim(Parcels)[1]
  Parcels$price <- round(exp(PreFloyd$LNPRICE), -2)
  Parcels$priceIni <- Parcels$price
  Parcels$OnSale <- FALSE
  Parcels$TonMarket <- 0
  if (!is.null(PreFloyd$dflood)) {Parcels$pflood <- Parcels$dflood*0.01}
  if (!is.null(PreFloyd$dflood100)&
      !is.null(PreFloyd$dflood500)) {Parcels$pflood <- Parcels$dflood100*0.01+Parcels$dflood500*0.002}
  Parcels$tradershere <- 0
  # calculate insurance rates:
  ins100 <- 0.0098*Parcels$price+30+75 # ICC premium equals USD 75 and Federal policy fee USD 30 for 100-year floodplain
  ins100[Parcels$price>50000] <- 0.0008*Parcels$price[Parcels$price>50000]+450+30+75
  ins100[Parcels$price>250000] <- 650+30+75
  ins500 <- 0.0058*Parcels$price+30+6 # ICC premium equals USD 6 and Federal policy fee USD 30 for 500-year floodplain
  ins500[Parcels$price>50000] <- 0.0014*Parcels$price[Parcels$price>50000]+220+30+6
  ins500[Parcels$price>250000] <- 570+30+6
  Parcels$FloodIns <- 0
  if (Insurance) {Parcels$FloodIns[Parcels$pflood==0.01] <- ins100[Parcels$pflood==0.01]}  
  if (Insurance) {Parcels$FloodIns[Parcels$pflood==0.002] <- ins500[Parcels$pflood==0.002]}  
  # Select neighbours for the parcels and see if any of them are in the flood zone
  TEMP <- dist(coordinates(Parcels))
  TEMP <- as.matrix(TEMP)
  TEMP <- TEMP < NeighbourR & TEMP > 0
  for (i in Parcels$ID) {
    nghbrs <- TEMP[,i]
    if (any(nghbrs)) {
      Parcels$Nghbrs.100FZ[i] <- any(Parcels$dflood100[nghbrs]==1)
      Parcels$Nghbrs.500FZ[i] <- any(Parcels$dflood500[nghbrs]==1)
    } else {
      Parcels$Nghbrs.100FZ[i] <- 0
      Parcels$Nghbrs.500FZ[i] <- 0
    }
  }
  remove(TEMP)
  
# add households to the parcels. each parcel will be occupied by one household
# all the relevant agent attributes are attached to the households here as well (similar to the 'turtles own' function in NetLogo)
  Ntraders <- NParcels
  Traders <- data.frame(Parcels$ID)
  names(Traders) <- "PID"
  Traders$Utility <- 0
  Traders$House.PFLOOD <- Parcels$pflood
  Traders$MarketStatus <- "Traded"
  Traders$MyPrice <- Parcels$price
  Traders$Income <- exp(5.4+0.474*log(Traders$MyPrice))
  Traders$Budget <- round(exp(4.95765+0.63445*log(Traders$Income)+rnorm(Ntraders,sd=0.7701126)),-3)
  Traders$minBudget <- runif(Ntraders,.5,0.8421) * Traders$Budget
  Traders$Refused <- 0
  Traders$aDelta <- round(Traders$MyPrice*kH/kY)
  Traders$CORinsurance <- FALSE
  Traders$CORfloodrisk <- FALSE
    # Risk perception and risk behaviour variables from survey:
  Traders$Exp.Flood <- runif(Ntraders) < 0.252                                                  # Number of people that have experienced a flooding in their town
  Traders$Exp.Damage <- runif(Ntraders) < 0.069/0.252 & Traders$Exp.Flood                       # Number of people that have had damage from flooding before
  Traders$Learning.Not <- runif(Ntraders) < 0.263/(1-0.252) & !Traders$Exp.Flood
  Traders$Learning.NewsMedia <- runif(Ntraders) < 0.278/(1-0.263) & !Traders$Learning.Not
  Traders$Learning.Nghbrs <- runif(Ntraders) < 0.292/(1-0.263) & !Traders$Learning.Not
  Traders$LikelyDamage <- sample(SurveyDATA$LikelyDamage, Ntraders, replace=TRUE)
  for (i in 1:nrow(LikelyDamage.Sel)) {
    rules <- Traders$House.PFLOOD==LikelyDamage.Sel$House.PFLOOD[i] &
      Traders$Learning.Nghbrs==LikelyDamage.Sel$Learning.Nghbrs[i] &
      Traders$Exp.Damage==LikelyDamage.Sel$Exp.Damage2[i]
    LD.sample <- sample(1:4,sum(rules),replace=TRUE,prob=LikelyDamage.Sel[i,c("1","2","3","4")])
    Traders$LikelyDamage[rules] <- LD.sample
  }
  Traders$Costs.FloodIns <- sample(SurveyDATA$Costs.FloodIns, Ntraders, replace=TRUE)
  Traders$Effect.FloodIns <- sample(SurveyDATA$Effect.FloodIns, Ntraders, replace=TRUE)
  Traders$Costs.MovingOut <- sample(SurveyDATA$Costs.MovingOut, Ntraders, replace=TRUE)
  Traders$Effect.MovingOut <- sample(SurveyDATA$Effect.MovingOut, Ntraders, replace=TRUE)
  Traders$Prob.FloodIns <- 0.145+0.07*Traders$Effect.FloodIns-0.053*Traders$Costs.FloodIns+0.126*Traders$LikelyDamage+22.4*Traders$House.PFLOOD
  Traders$FloodIns <- runif(Ntraders) < Traders$Prob.FloodIns
  Traders$FEARscore <- round(sample(SurveyDATA$FEARscore, Ntraders, replace=TRUE),1)          # People's fear towards flooding
  for (i in 1:nrow(FEARscore)) {
    rules <- Traders$Learning.Not==FEARscore$Learning.Not[i] &
      Traders$Learning.NewsMedia==FEARscore$Learning.NewsMedia[i] &
      Traders$Exp.Damage==FEARscore$Learning.NewsMedia[i] 
    FS.sample <- sample(as.numeric(names(FEARscore)[5:ncol(FEARscore)]),sum(rules),
                        replace=TRUE,prob=FEARscore[i,5:ncol(FEARscore)])
    Traders$FEARscore[rules] <- FS.sample
  }
  Traders$FloodIssue <- runif(Ntraders) < Traders$Learning.Not*-.18+Traders$Exp.Flood*.18+.416 
  Traders$BEHAVscore <- round(sample(SurveyDATA$BEHAVscore, Ntraders, replace=TRUE),1)                 # The tendency for buyers to avoid the flood zone.
  for (i in 1:nrow(BEHAVscore)) {
    rules <- Traders$FEARscore==BEHAVscore$FEARscore[i] &
      Traders$FloodIssue==BEHAVscore$Flood.issue[i]
    BS.sample <- sample(as.numeric(names(BEHAVscore)[4:ncol(BEHAVscore)]),sum(rules),
                        replace=TRUE,prob=BEHAVscore[i,4:ncol(BEHAVscore)])
    Traders$BEHAVscore[rules] <- BS.sample
  }
  Traders$AVOID100FZ <- runif(Ntraders) < Traders$BEHAVscore*0.09389-0.09090                    # a number of buyers avoid the 100-year flood zone
  Traders$AVOID500FZ[Traders$AVOID100FZ] <- runif(sum(Traders$AVOID100FZ)) < 0.565              # a number of buyers that avoid the 100-year flood zone will also avoid the 500-year flood zone, derived by Bayesian inference
  Traders$AVOID500FZ[!Traders$AVOID100FZ] <- runif(sum(!Traders$AVOID100FZ)) < 0.09             # a number of buyers that do not avoid the 100-year flood zone will avoid the 500-year flood zone, derived by Bayesian inference
  Traders$LikelyMoveOut <- 0.02*Traders$FEARscore+0.16*Traders$Exp.Damage-0.034
  Traders$Mortgage <- sample(SurveyDATA$Mortgage[SurveyDATA$MarketStatus=="Buyer"],Ntraders,replace=TRUE)
  Traders$Financed <- round((runif(Ntraders)*100^3)^(1/3))
  Traders$Financed[Traders$Mortgage=="No"] <- 0
  Traders$yM <- sample(1:yM,Ntraders,replace=TRUE)
  Traders$Debt <- round(Traders$MyPrice*Traders$Financed/100*(1-Traders$yM/yM))
  Traders$SellerStatus <- "Inactive"
  
  Traders <- cbind(Traders, BuyerPrefs[sample(1:dim(BuyerPrefs)[1],Ntraders, replace=TRUE),])
  
  print("Successfully loaded landscape")
  
  #write tables - these tables will keep track of all the sales and the relevant attributes of the people that buy and sell properties.
  setwd(tempdir())
  SALES <- data.frame("Tick", "ID", "SalesPrice", "BuyerBudget")
  names(SALES) <- c("Tick", "ID", "SalesPrice", "BuyerBudget")
  SALES <- SALES[-1,]
  write.table(SALES, "Trade results.csv", sep=",")
  TRADES <- data.frame("Tick", "ID", "AskPrice", "BidPrice", " NofBuyers", "TradeCode")
  names(TRADES) <- c("Tick", "ID", "AskPrice", "BidPrice", " NofBuyers", "TradeCode")
  TRADES <- TRADES[-1,]
  write.table(TRADES, "All trades.csv", sep=",")
  setwd(dir)
  HOMEOWNERINFO <- cbind(rep(TICK,NParcels),Traders[Traders$MarketStatus!="Buyer",])
  names(HOMEOWNERINFO)[1] <- "TICK"
  # reduce the size of HOMEOWNERINFO dataset - because it is a very large dataset
  HOMEOWNERINFO <- HOMEOWNERINFO[c("TICK","PID","House.PFLOOD","Income","Exp.Flood","Exp.Damage","FloodIns","FEARscore","Mortgage","Debt","SellerStatus")]
  HOMEOWNERINFO <- HOMEOWNERINFO[HOMEOWNERINFO$TICK>90,]
  HOMEOWNERINFO$Debt <- round(HOMEOWNERINFO$Debt,-2)
  HOMEOWNERINFO$Income <- round(HOMEOWNERINFO$Income,-2)
  
  # calculate the average rent based on the average price of the properties
  avRent <- mean(Parcels$price) * kH * aDelta / kY
  totalPV <- sum(Parcels$price)
  countNonTrades <- 0
  yearCount <- 0
  # things will change in the scenarios when it is the year of the flood
  FloodYear <- FALSE
  # things will change in the scenarios when floods are covered by news media
  MediaCoverage <- FALSE
  NewBuyerOS <- 0
  
  #-----
  # GO - code for the dynamics of the model
  #-----
  
  for (TICK in 1:SIMlength) {
    #Parcels on sale update their time on market
    Parcels$TonMarket[Parcels$OnSale] <- Parcels$TonMarket[Parcels$OnSale] + 1
    
    # Flood event #
    if(substr(Scenario,1,5)=="Flood" & (TICK-100)/kY == 1) {FloodYear <- TRUE} else {
      if(substr(Scenario,1,7)=="Extreme" &
         ((TICK-100)/kY == 1 | (TICK-100)/kY == 5)) {FloodYear <- TRUE} else {FloodYear <- FALSE}}
    
    # Media event #
    if(FloodYear) {MediaCoverage <- TRUE} else {
      if(substr(Scenario,nchar(Scenario)-3,nchar(Scenario))=="edia" & TICK >= 100 &
         ((TICK-100)/kY)%%1 == 0) {MediaCoverage <- TRUE} else {MediaCoverage <- FALSE}  
    }
    
    ####################################
    ### Update all trader variables  ###
    ####################################
    
    Ntraders <- nrow(Traders)
    Traders$Debt[Traders$yM!=30] <- round(Traders$Debt[Traders$yM!=30]*(1-1/((yM-Traders$yM[Traders$yM!=30])*kY)))
    Traders$yM[Traders$Mortgage=="Yes"] <- Traders$yM[Traders$Mortgage=="Yes"] + 1/kY
    Traders$yM[Traders$yM>30] <- 30
    # every half year there is a 10% chance that people forget about flooding in news media
    Traders$Learning.NewsMedia[Traders$Learning.NewsMedia] <- runif(sum(Traders$Learning.NewsMedia)) < 0.9
    # 80% of the people are informed about flood risk through media if there is media coverage
    if (MediaCoverage) {Traders$Learning.NewsMedia[runif(nrow(Traders))<0.8] <- TRUE}
    # In a flood year everyone has experience with the flooding. All people living in the flood zone have damage.
    if (FloodYear) {
      Traders$Exp.Flood <- TRUE
      Traders$Exp.Damage[Parcels$ID][Parcels$pflood>0] <- TRUE
      Traders$Learning.NewsMedia <- TRUE
      Traders$Learning.Nghbrs[Parcels$ID][Parcels$Nghbrs.100FZ==1|Parcels$Nghbrs.500FZ==1] <- TRUE
    }
    Traders$Learning.Not[Traders$Learning.NewsMedia|Traders$Learning.Nghbrs|Traders$Exp.Flood] <- FALSE
    # Likelihood of damage of homeowners gets updated every half year
    LDnew <- sample(SurveyDATA$LikelyDamage, NParcels, replace=TRUE)
    for (i in 1:nrow(LikelyDamage.Sel)) {
      rules <- Traders$House.PFLOOD[Parcels$ID]==LikelyDamage.Sel$House.PFLOOD[i] &
        Traders$Learning.Nghbrs[Parcels$ID]==LikelyDamage.Sel$Learning.Nghbrs[i] &
        Traders$Exp.Damage[Parcels$ID]==LikelyDamage.Sel$Exp.Damage2[i]
      LD.sample <- sample(1:4,sum(rules),replace=TRUE,prob=LikelyDamage.Sel[i,c("1","2","3","4")])
      LDnew[rules] <- LD.sample
    }
    LDold <- Traders$LikelyDamage[Parcels$ID]
    LDnew[abs(LDnew-LDold)>1] <- LDold[abs(LDnew-LDold)>1]    # Perceived likelihood of damage cannot change by more than 1 point
    Traders$LikelyDamage[Parcels$ID] <- LDnew
    # Fear score of all households gets updated every half year
    FSnew <- round(sample(SurveyDATA$FEARscore, Ntraders, replace=TRUE),1)          # People's fear towards flooding
    for (i in 1:nrow(FEARscore)) {
      rules <- Traders$Learning.Not==FEARscore$Learning.Not[i] &
        Traders$Learning.NewsMedia==FEARscore$Learning.NewsMedia[i] &
        Traders$Exp.Damage==FEARscore$Learning.NewsMedia[i] 
      FS.sample <- sample(as.numeric(names(FEARscore)[5:ncol(FEARscore)]),sum(rules),
                          replace=TRUE,prob=FEARscore[i,5:ncol(FEARscore)])
      FSnew[rules] <- FS.sample
    }
    FSold <- Traders$FEARscore
    FSnew[abs(FSnew-FSold)>1] <- FSold[abs(FSnew-FSold)>1]    # Fear score cannot change by more than 1 point
    Traders$FEARscore <- FSnew
    # Buyer's tendency to avoid the flood zone gets updated
    Nbuyers <- sum(Traders$MarketStatus=="Buyer")
    if (Nbuyers > 0) {
      # first a number of buyers will reconsider if they take flood risk into account in their decisions
      FInew <- runif(Nbuyers) < Traders$Learning.Not[Traders$MarketStatus=="Buyer"]*-.18+Traders$Exp.Flood[Traders$MarketStatus=="Buyer"]*.18+.416 
      FIold <- Traders$FloodIssue[Traders$MarketStatus=="Buyer"]
      Nupdate <- sum(FInew&!FIold)-sum(!FInew&FIold)
      FIold[FInew&!FIold] <- runif(sum(FInew&!FIold))<Nupdate/sum(FInew&!FIold)
      FInew<-FIold
      Traders$FloodIssue[Traders$MarketStatus=="Buyer"] <- FInew
      # then their tendency to avoid the flood zone gets updated
      BSnew <- round(sample(SurveyDATA$BEHAVscore, Nbuyers, replace=TRUE),1)     # The tendency for buyers to avoid the flood zone.
      for (i in 1:nrow(BEHAVscore)) {
        rules <- Traders$FEARscore[Traders$MarketStatus=="Buyer"]==BEHAVscore$FEARscore[i] &
          Traders$FloodIssue[Traders$MarketStatus=="Buyer"]==BEHAVscore$Flood.issue[i]
        BS.sample <- sample(as.numeric(names(BEHAVscore)[4:ncol(BEHAVscore)]),sum(rules),
                            replace=TRUE,prob=BEHAVscore[i,4:ncol(BEHAVscore)])
        BSnew[rules] <- BS.sample
      }
      BSold <- Traders$BEHAVscore[Traders$MarketStatus=="Buyer"]
      BSnew[abs(BSnew-BSold)>1] <- BSold[abs(BSnew-BSold)>1]    # Behaviour score cannot change by more than 1 point
      Traders$BEHAVscore[Traders$MarketStatus=="Buyer"] <- BSnew
      # Finally, they update if they actively search for properties outside the flood zone
      AV100FZnew <- runif(Nbuyers) < Traders$BEHAVscore[Traders$MarketStatus=="Buyer"]*0.09389-0.09090                    # a number of buyers avoid the 100-year flood zone
      AV100FZold <- Traders$AVOID100FZ[Traders$MarketStatus=="Buyer"]
      Nupdate <- sum(AV100FZnew&!AV100FZold)-sum(!AV100FZnew&AV100FZold)
      AV100FZold[AV100FZnew&!AV100FZold] <- runif(sum(AV100FZnew&!AV100FZold))<Nupdate/sum(AV100FZnew&!AV100FZold)
      AV100FZnew<-AV100FZold
      Traders$AVOID100FZ[Traders$MarketStatus=="Buyer"] <- AV100FZnew
      Traders$AVOID500FZ[Traders$AVOID100FZ&Traders$MarketStatus=="Buyer"] <- runif(sum(Traders$AVOID100FZ[Traders$MarketStatus=="Buyer"])) < 0.565              # a number of buyers that avoid the 100-year flood zone will also avoid the 500-year flood zone, derived by Bayesian inference
      Traders$AVOID500FZ[!Traders$AVOID100FZ&Traders$MarketStatus=="Buyer"] <- runif(sum(!Traders$AVOID100FZ[Traders$MarketStatus=="Buyer"])) < 0.09             # a number of buyers that do not avoid the 100-year flood zone will avoid the 500-year flood zone, derived by Bayesian inference
    }
    # Homeowners update their probability to abandon their risk-prone property
    Traders$LikelyMoveOut[Parcels$ID] <- 0.02*Traders$FEARscore[Parcels$ID]+0.16*Traders$Exp.Damage[Parcels$ID]-0.034
    Traders$SellerStatus[Traders$MarketStatus=="Buyer"] <- "None"
    Traders$SellerStatus[Traders$MarketStatus=="Traded"] <- "Inactive"
    
    ####################################
    ### Done updating RP variables   ###
    ####################################

    #Buyers and sellers who have been on the market for more than 2 years will leave the market
    Traders$SellerStatus[Traders$Refused>4 & Traders$MarketStatus=="Seller"] <- "Failed to sell"
    Traders$MarketStatus[Traders$Refused>4 & Traders$MarketStatus=="Seller"] <- "Traded"
    Traders <- Traders[!(Traders$Refused>4 & Traders$MarketStatus=="Buyer"),]
    Parcels$OnSale[Parcels$TonMarket>4] <- FALSE
    Traders$Refused[Traders$MarketStatus=="Traded"] <- 0
    Parcels$TonMarket[!Parcels$OnSale] <- 0
    
    #Update rent
    avRent <- mean(Parcels$price) * kH * aDelta / kY  
    # Create new sellers and form ask price
    NewSales <- round(rnorm(1,fraction_on_sale*NParcels,sd_fraction_on_sale*NParcels) / kY)
    ID <- sample(Parcels$ID[!Parcels$OnSale], NewSales)        # gives the ID of new properties on sale. this command excludes properties that were already on sale
    # A number of sellers in the flood zone will choose to evade the hazard
    evadeID <- Parcels$ID[runif(NParcels) < Traders$LikelyMoveOut[1:NParcels] & 
                            Parcels$pflood > 0 & !Parcels$OnSale]
    ID <- c(ID,evadeID)
    ID <- unique(ID)
    Parcels$OnSale[ID] <- TRUE
    
    #sellers who have been refused twice will lower their price
    ID <- Parcels$ID[Parcels$TonMarket%%2 == 0 & Parcels$OnSale]   # gives the ID of properties that are either new on the market or that have been refused twice.
    
    NonSale <- dim(Parcels[Parcels$OnSale,])[1]
    Traders$MarketStatus[Traders$PID %in% ID] <- "Seller"
    Traders$SellerStatus[Traders$MarketStatus=="Seller"] <- "Attempting to sell"
    Traders$SellerStatus[Traders$PID %in% evadeID] <- "Evading hazard"
    # Sellers that are new on the market will decide if they take flood risk and insurances into account when setting the price. For now, this does not need to be updated for sellers who are already on the market.
    # Perhaps this can be updated at a later stage. For example when flooding occurs and people need to take flood risk into account.
    Traders$CORinsurance[Traders$MarketStatus=="Seller" & Traders$Refused==0] <- FALSE
    Traders$CORfloodrisk[Traders$MarketStatus=="Seller" & Traders$Refused==0] <- TRUE
    CORinsurance <- Traders$CORinsurance[Traders$PID %in% ID]
    CORfloodrisk <- Traders$CORfloodrisk[Traders$PID %in% ID]
    
    # # # RUN HEDONICS  # # #
    
    # function to determine the ask price for all properties on sale on the bases of previous sales
    # it is based on hedonic analysis of property sales and spatial interpolation (Kriging) of the residuals (de Koning et al. 2016) https://research.utwente.nl/en/publications/improved-methods-for-predicting-property-prices-in-hazard-prone-d
    getPrice <- function(ID, CORfloodrisk,Expectations="Dynamic") {
      
      # Select properties on sale
      subset <- ID
      PreFloyd.sub <- PreFloyd[subset,]
      PRICE <- rep(0, length(ID))
      
      # Select the previous sales and run the hedonic regression
      setwd(tempdir())
      Sales <- read.table("Trade results.csv", sep=",", header=TRUE)
      setwd(dir)
      first <- function() {prevsales <- Sales$ID[Sales$Tick == max(Sales$Tick)]
      SalesPrice <- Sales$SalesPrice[Sales$Tick == max(Sales$Tick)]
      SalesPrice[SalesPrice==0] <- 1000
      SalesPrice <- log(SalesPrice)
      PreFloyd.rest <- PreFloyd[prevsales,]
      PreFloyd.rest$LNPRICE <- SalesPrice
      return(PreFloyd.rest)}
      runhedonics.flooddummy <- function() {
        
        if (!is.null(PreFloyd$dflood)) {
          df <- data.frame(PreFloyd.rest$LNPRICE,PreFloyd.rest$age,log(PreFloyd.rest$TOTALSQFT),PreFloyd.rest$LNACRES,PreFloyd.rest$bedrm,PreFloyd.rest$dflood)
          names(df) <- c("LNPRICE", "age", "log(TOTALSQFT)", "LNACRES", "bedrm", "dflood")
        }
        
        if (!is.null(PreFloyd$dflood100)&
            !is.null(PreFloyd$dflood500)) {
          df <- data.frame(PreFloyd.rest$LNPRICE,PreFloyd.rest$age,log(PreFloyd.rest$TOTALSQFT),PreFloyd.rest$LNACRES,PreFloyd.rest$bedrm,PreFloyd.rest$dflood100,PreFloyd.rest$dflood500)
          names(df) <- c("LNPRICE", "age", "log(TOTALSQFT)", "LNACRES", "bedrm", "dflood100", "dflood500")
        }
        
        hedonics <- lm(LNPRICE~., df)
        HED.coef <- coef(hedonics)
        HED.coef[is.na(HED.coef)] <- 0
        nvars <- length(HED.coef)
        if (nvars == 6) exp.sign <- c(-1,1,1,1,-1) else exp.sign <- c(-1,1,1,1,-1,-1)
        # expected sign?
        right.sign <- exp.sign * HED.coef[2:nvars] > 0
        df <- df[,c(TRUE,right.sign)]
        # adjust only the coefficients for significant variables with the right expected sign
        HED.coef[c(TRUE,right.sign)] <- coef(lm(LNPRICE~., df))
        # set other coefficients to 0
        HED.coef[c(FALSE,!right.sign)] <- 0
        return(HED.coef)}
      runhedonics.nodummy <- function() {df <- data.frame(PreFloyd.rest$LNPRICE,PreFloyd.rest$age,log(PreFloyd.rest$TOTALSQFT),PreFloyd.rest$LNACRES,PreFloyd.rest$bedrm)
      names(df) <- c("LNPRICE", "age", "log(TOTALSQFT)", "LNACRES", "bedrm")
      hedonics <- lm(LNPRICE~., df)
      HED.coef <- coef(hedonics)
      HED.coef[is.na(HED.coef)] <- 0
      exp.sign <- c(-1,1,1,1)
      # expected sign?
      right.sign <- exp.sign * HED.coef[2:5] > 0
      df <- df[,c(TRUE,right.sign)]
      # adjust only the coefficients for significant variables with the right expected sign
      HED.coef[c(TRUE,right.sign)] <- coef(lm(LNPRICE~., df))
      # set other coefficients to 0
      HED.coef[c(FALSE,!right.sign)] <- 0
      return(HED.coef)}
      if (dim(Sales)[1] > 0) PreFloyd.rest <- try(first()) else PreFloyd.rest <- PreFloyd[-subset,]
      if (dim(Sales)[1] > 0) {if (Expectations=="Dynamic") HED.coef.flooddummy <- try(runhedonics.flooddummy())}
      if (dim(Sales)[1] > 0) {if (Expectations=="Dynamic") HED.coef.nodummy <- try(runhedonics.nodummy())}
      
      # for debugging:
      print(" ")
      print(" ")
      print("Hedonic coefficients for realtors who take the flood zone into account:")
      print(" ")
      if(exists("HED.coef.flooddummy"))print(HED.coef.flooddummy)
      print(" ")
      print(" ")
      print("Hedonic coefficients for all other realtors:")
      print(" ")
      if(exists("HED.coef.nodummy")){print(HED.coef.nodummy)}
      print(" ")
      print(" ")
      # end debugging
      
      ###########################################################
      # part I: for realtors that take flood risk into account  #
      ###########################################################
      if(sum(CORfloodrisk)!=0) {
        if (!is.null(PreFloyd$dflood)) {
          PREDICT <- HED.coef.flooddummy[1] + 
            HED.coef.flooddummy[2] * PreFloyd.rest$age +
            HED.coef.flooddummy[3] * log(PreFloyd.rest$TOTALSQFT) +
            HED.coef.flooddummy[4] * PreFloyd.rest$LNACRES +
            HED.coef.flooddummy[5] * PreFloyd.rest$bedrm +
            HED.coef.flooddummy[6] * PreFloyd.rest$dflood 
        }    # when there is only one dummy for flood zone
        if (!is.null(PreFloyd$dflood100)&
            !is.null(PreFloyd$dflood500)) {
          PREDICT <- HED.coef.flooddummy[1] + 
            HED.coef.flooddummy[2] * PreFloyd.rest$age +
            HED.coef.flooddummy[3] * log(PreFloyd.rest$TOTALSQFT) +
            HED.coef.flooddummy[4] * PreFloyd.rest$LNACRES +
            HED.coef.flooddummy[5] * PreFloyd.rest$bedrm +
            HED.coef.flooddummy[6] * PreFloyd.rest$dflood100 +
            HED.coef.flooddummy[7] * PreFloyd.rest$dflood500 
        }  # when there is a seperation between 100-year and 500-year flood zone
        PreFloyd.rest$RESID <- PreFloyd.rest$LNPRICE - PREDICT
        
        # Predict price of subset according to remaining set
        #coefficients
        PreFloyd.sub.flooddummy <- PreFloyd.sub[CORfloodrisk,]
        if (!is.null(PreFloyd$dflood)) {
          PREDICT <- HED.coef.flooddummy[1] + 
            HED.coef.flooddummy[2] * PreFloyd.sub.flooddummy$age +
            HED.coef.flooddummy[3] * log(PreFloyd.sub.flooddummy$TOTALSQFT) +
            HED.coef.flooddummy[4] * PreFloyd.sub.flooddummy$LNACRES +
            HED.coef.flooddummy[5] * PreFloyd.sub.flooddummy$bedrm +
            HED.coef.flooddummy[6] * PreFloyd.sub.flooddummy$dflood
        }
        if (!is.null(PreFloyd$dflood100)&
            !is.null(PreFloyd$dflood500)) {
          PREDICT <- HED.coef.flooddummy[1] + 
            HED.coef.flooddummy[2] * PreFloyd.sub.flooddummy$age +
            HED.coef.flooddummy[3] * log(PreFloyd.sub.flooddummy$TOTALSQFT) +
            HED.coef.flooddummy[4] * PreFloyd.sub.flooddummy$LNACRES +
            HED.coef.flooddummy[5] * PreFloyd.sub.flooddummy$bedrm +
            HED.coef.flooddummy[6] * PreFloyd.sub.flooddummy$dflood100 +
            HED.coef.flooddummy[7] * PreFloyd.sub.flooddummy$dflood500
        }  
        #kriging
        RESID <- krige(RESID~1, PreFloyd.rest, PreFloyd.sub.flooddummy, model = vgm.fit.RESID, nmax=10)
        RESID <- RESID$var1.pred
        #expected sales price
        PRICE[CORfloodrisk] <- PREDICT + RESID}
      
      ###################################################################
      # part II: for realtors that do not take flood risk into account  #
      ###################################################################
      if(sum(!CORfloodrisk)!=0) {
        PREDICT <- HED.coef.nodummy[1] + 
          HED.coef.nodummy[2] * PreFloyd.rest$age +
          HED.coef.nodummy[3] * log(PreFloyd.rest$TOTALSQFT) +
          HED.coef.nodummy[4] * PreFloyd.rest$LNACRES +
          HED.coef.nodummy[5] * PreFloyd.rest$bedrm
        PreFloyd.rest$RESID <- PreFloyd.rest$LNPRICE - PREDICT
        
        # Predict price of subset according to remaining set
        #coefficients
        PreFloyd.sub.nodummy <- PreFloyd.sub[!CORfloodrisk,]
        PREDICT <- HED.coef.nodummy[1] + 
          HED.coef.nodummy[2] * PreFloyd.sub.nodummy$age +
          HED.coef.nodummy[3] * log(PreFloyd.sub.nodummy$TOTALSQFT) +
          HED.coef.nodummy[4] * PreFloyd.sub.nodummy$LNACRES +
          HED.coef.nodummy[5] * PreFloyd.sub.nodummy$bedrm
        #kriging
        RESID <- krige(RESID~1, PreFloyd.rest, PreFloyd.sub.nodummy, model = vgm.fit.RESID, nmax=10)
        RESID <- RESID$var1.pred
        #expected sales price
        PRICE[!CORfloodrisk] <- PREDICT + RESID 
      }
      
      return(PRICE)
      
    }
    # function to adjust the prices on the demand for similar properties in previous time steps
    PriceAdj2 <- function(ID, TICK,alpha) {
      if(TICK>1){
        setwd(tempdir())
        Trades <- read.table("All trades.csv", sep=",", header=TRUE)
        setwd(dir)
        Trades <- Trades[Trades$Tick==max(Trades$Tick),]
        Trades$LNSQFT <- log(PreFloyd$TOTALSQFT[Trades$ID])
        Trades$LNACRES <- PreFloyd$LNACRES[Trades$ID]
        Trades$BEDRM <- PreFloyd$bedrm[Trades$ID]
        Trades$AGE <- PreFloyd$age[Trades$ID]
        if (!is.null(PreFloyd$dflood)) {PreFloyd$pflood <- PreFloyd$dflood*0.01}
        if (!is.null(PreFloyd$dflood100)&
            !is.null(PreFloyd$dflood500)) {PreFloyd$pflood <- PreFloyd$dflood100*0.01+PreFloyd$dflood500*0.002}
        Trades$PFLOOD <- PreFloyd$pflood[Trades$ID]
        #Trades$AMEN <- PreFloyd$RESID[Trades$ID]
        coords <- coordinates(PreFloyd)
        correct <- rep(NA, length(ID))
        for (i in 1:length(ID)) {
          lnsqft <- log(PreFloyd$TOTALSQFT[ID[i]])
          lnacres <- PreFloyd$LNACRES[ID[i]]
          age <- PreFloyd$age[ID[i]]
          pflood <- PreFloyd$pflood[ID[i]]
          amen <- coords[ID[i],]
          
          CONDsqft <- Trades$LNSQFT >= lnsqft - .2 & Trades$LNSQFT <= lnsqft + .2
          CONDacres <- Trades$LNACRES >= lnacres - .5 & Trades$LNACRES <= lnacres + .5
          CONDage <- Trades$AGE >= age - 2 & Trades$AGE <= age + 2
          CONDpflood <- Trades$PFLOOD == pflood
          EucDist <- sqrt((amen[1]-coords[Trades$ID,1])^2+(amen[2]-coords[Trades$ID,2])^2)
          CONDamen <- EucDist < 4000
          
          Trades.sqft <- Trades[CONDsqft,]
          Trades.acres <- Trades[CONDacres,]
          Trades.age <- Trades[CONDage,]
          Trades.pflood <- Trades[CONDpflood,]
          Trades.amen <- Trades[CONDamen,]
          
          Nobs.sqft <- mean(Trades.sqft$NofBuyers)
          Nobs.acres <- mean(Trades.acres$NofBuyers)
          Nobs.age <- mean(Trades.age$NofBuyers)
          Nobs.pflood <- mean(Trades.pflood$NofBuyers)
          Nobs.amen <- mean(Trades.amen$NofBuyers)
          Nexp <- mean(Trades$NofBuyers)
          
          CORRECT.sqft <- alpha*(Nobs.sqft/Nexp-1)
          if(is.na(CORRECT.sqft)) {CORRECT.sqft <- 0}
          if(CORRECT.sqft > alpha) {CORRECT.sqft <- alpha}
          
          CORRECT.acres <- alpha*(Nobs.acres/Nexp-1)
          if(is.na(CORRECT.acres)) {CORRECT.acres <- 0}
          if(CORRECT.acres > alpha) {CORRECT.acres <- alpha}
          
          CORRECT.age <- alpha*(Nobs.age/Nexp-1)
          if(is.na(CORRECT.age)) {CORRECT.age <- 0}
          if(CORRECT.age > alpha) {CORRECT.age <- alpha}
          
          CORRECT.pflood <- alpha*(Nobs.pflood/Nexp-1)
          if(is.na(CORRECT.pflood)) {CORRECT.pflood <- 0}
          if(CORRECT.pflood > alpha) {CORRECT.pflood <- alpha}
          
          CORRECT.amen <- alpha*(Nobs.amen/Nexp-1)
          if(is.na(CORRECT.amen)) {CORRECT.amen <- 0}
          if(CORRECT.amen > alpha) {CORRECT.amen <- alpha}
          
          CORRECT <- CORRECT.sqft+CORRECT.acres+CORRECT.age+CORRECT.pflood+CORRECT.amen
          
          correct[i] <- CORRECT
        }
        return(correct)
      }else{return(0)}
    }
    
    rResult <- round(exp(as.numeric(getPrice(ID, CORfloodrisk,Expectations)+
                                      PriceAdj2(ID, TICK,alpha))),-2)
    
    # # # END of Hedonic pricing + kriging function # # #
    
    rResult[CORinsurance] <- rResult[CORinsurance] - (Parcels$FloodIns[ID]/kH)[CORinsurance] # lower the price to account for the costs of flood insurance
    #sellers who have been refused twice will lower their price
    rResult[Parcels$TonMarket[ID] > 0] <- rResult[Parcels$TonMarket[ID] > 0] * 0.96
    #sellers who had damage from the flood and sellers in the 100-year flood zone who experienced a flood will lower their price
    if (FloodYear) {
      rules100FZxEXP <- Traders$House.PFLOOD[Traders$PID %in% ID]==0.01 & Traders$Exp.Flood[Traders$PID %in% ID]
      rResult[rules100FZxEXP] <- rResult[rules100FZxEXP]*exp(EXPreduction)  # price reduction of 15% according to survey
      rulesDAMAGE <- Traders$Exp.Damage[Traders$PID %in% ID]
      rResult[rulesDAMAGE] <- rResult[rulesDAMAGE]*exp(DMGreduction)  # price reduction of 13% according to survey
    }
    rResult <- round(rResult, -2)
    rResult[rResult<=0] <- 1000 # ask price should always be higher than 0
    Parcels$price[ID] <- rResult
    Traders$MyPrice[Traders$PID %in% ID] <- rResult
    #sellers that cannot afford to sell their property will withdraw their sale
    withdraw <- Traders$MyPrice[Traders$PID %in% ID] < Traders$Debt[Traders$PID %in% ID]
    Parcels$OnSale[ID][withdraw] <- FALSE
    Traders$MarketStatus[Traders$PID %in% ID][withdraw] <- "Traded"
    Traders$SellerStatus[Traders$PID %in% ID][withdraw] <- "Cannot afford selling"
    ID <- ID[!withdraw]
    NonSale <- dim(Parcels[Parcels$OnSale,])[1]
    
    #update graphics
    if(UPDATEGRAPH){
      plot(Parcels[!Parcels$OnSale,], pch=15, col="dark red", add=TRUE, cex=0.3)
      plot(Parcels[Parcels$OnSale,], pch=15, col="green", add=TRUE, cex=.3)
    }
    Trades <- data.frame(rep(TICK, NonSale), Parcels$ID[Parcels$OnSale], Parcels$price[Parcels$OnSale])
    names(Trades) <- c("Tick", "ID", "AskPrice")
    Trades$BidPrice <- 0
    Trades$NofBuyers <- 0
    Trades$TradeCode <- 0
    
    #debugging
    if (dim(Traders[Traders$MarketStatus=="Seller",])[1] != dim(Parcels[Parcels$OnSale,])[1]) {
      print(paste("Interrupted at T =", TICK))
      break}
    
  # Create new buyers that enter the market and will search for properties that match their budget and preferences.
  # attach attributes to these buyers
    if (NofBuyers == "Equal to sellers") {NewBuyers <- round(rnorm(1,fraction_on_sale*NParcels,sd_fraction_on_sale*NParcels) / kY)}
    NewBuyers <- NewBuyers-NewBuyerOS      # Part of the new buyers are already there: were sellers in the previous time step
    Buyers <- Traders[1:NewBuyers,]        # Buyers is a dataset with new buyers only
    Buyers$PID <- 0
    Buyers$House.PFLOOD <- 0
    Buyers$MarketStatus <- "Buyer"
    Buyers$MyPrice <- 0
    Buyers$Income <- sample(income, NewBuyers)
    Buyers$Budget <- round(exp(4.95765+0.63445*log(Buyers$Income)+rnorm(NewBuyers,sd=0.7701126)),-3)
    Buyers$minBudget <- runif(NewBuyers,.5,0.8421) * Buyers$Budget  # minimum % of their budget that buyers would like to spend, derived empirically
    # Risk perception and risk behaviour variables from survey:
    Buyers$Exp.Flood <- runif(NewBuyers) < 0.252                                                  # Number of people that have experienced a flooding in their town
    Buyers$Exp.Damage <- runif(NewBuyers) < 0.069/0.252 & Buyers$Exp.Flood                       # Number of people that have had damage from flooding before
    Buyers$Learning.Not <- runif(NewBuyers) < 0.263/(1-0.252) & !Buyers$Exp.Flood
    Buyers$Learning.NewsMedia <- runif(NewBuyers) < 0.278/(1-0.263) & !Buyers$Learning.Not
    Buyers$Learning.Nghbrs <- runif(NewBuyers) < 0.292/(1-0.263) & !Buyers$Learning.Not
    Buyers$LikelyDamage <- sample(SurveyDATA$LikelyDamage, NewBuyers, replace=TRUE)
    Buyers$Costs.FloodIns <- sample(SurveyDATA$Costs.FloodIns, NewBuyers, replace=TRUE)
    Buyers$Effect.FloodIns <- sample(SurveyDATA$Effect.FloodIns, NewBuyers, replace=TRUE)
    Buyers$Costs.MovingOut <- sample(SurveyDATA$Costs.MovingOut, NewBuyers, replace=TRUE)
    Buyers$Effect.MovingOut <- sample(SurveyDATA$Effect.MovingOut, NewBuyers, replace=TRUE)
    Buyers$Prob.FloodIns <- 0.145+0.07*Buyers$Effect.FloodIns-0.053*Buyers$Costs.FloodIns+0.126*Buyers$LikelyDamage
    Buyers$FloodIns <- runif(NewBuyers) < Buyers$Prob.FloodIns
    Buyers$FEARscore <- round(sample(SurveyDATA$FEARscore, NewBuyers, replace=TRUE),1)          # People's fear towards flooding
    for (i in 1:nrow(FEARscore)) {
      rules <- Buyers$Learning.Not==FEARscore$Learning.Not[i] &
        Buyers$Learning.NewsMedia==FEARscore$Learning.NewsMedia[i] &
        Buyers$Exp.Damage==FEARscore$Learning.NewsMedia[i] 
      FS.sample <- sample(as.numeric(names(FEARscore)[5:ncol(FEARscore)]),sum(rules),
                          replace=TRUE,prob=FEARscore[i,5:ncol(FEARscore)])
      Buyers$FEARscore[rules] <- FS.sample
    }
    Buyers$FloodIssue <- runif(NewBuyers) < Buyers$Learning.Not*-.18+Buyers$Exp.Flood*.18+.416 
    Buyers$BEHAVscore <- round(sample(SurveyDATA$BEHAVscore, NewBuyers, replace=TRUE),1)         # The tendency for buyers to avoid the flood zone.
    for (i in 1:nrow(BEHAVscore)) {
      rules <- Buyers$FEARscore==BEHAVscore$FEARscore[i] &
               Buyers$FloodIssue==BEHAVscore$Flood.issue[i]
      BS.sample <- sample(as.numeric(names(BEHAVscore)[4:ncol(BEHAVscore)]),sum(rules),
                          replace=TRUE,prob=BEHAVscore[i,4:ncol(BEHAVscore)])
      Buyers$BEHAVscore[rules] <- BS.sample
    }
    Buyers$AVOID100FZ <- runif(NewBuyers) < Buyers$BEHAVscore*0.09389-0.09090                    # a number of buyers avoid the 100-year flood zone
    Buyers$AVOID500FZ[Buyers$AVOID100FZ] <- runif(sum(Buyers$AVOID100FZ)) < 0.565              # a number of buyers that avoid the 100-year flood zone will also avoid the 500-year flood zone, derived by Bayesian inference
    Buyers$AVOID500FZ[!Buyers$AVOID100FZ] <- runif(sum(!Buyers$AVOID100FZ)) < 0.09             # a number of buyers that do not avoid the 100-year flood zone will avoid the 500-year flood zone, derived by Bayesian inference
    Buyers$Mortgage <- sample(SurveyDATA$Mortgage[SurveyDATA$MarketStatus=="Buyer"],NewBuyers,replace=TRUE)
    Buyers$Financed <- round((runif(NewBuyers)*100^3)^(1/3))
    Buyers$Financed[Buyers$Mortgage=="No"] <- 0
    Buyers$yM <- 0
    Buyers$Debt <- 0
    Buyers$SellerStatus <- "None"
    
    Buyers[,c("AGEnorm","SQFTnorm","ACRESnorm","BEDRMnorm", "RESIDnorm")] <- # set buyer preferences for housing attributes
      BuyerPrefs[sample(1:dim(BuyerPrefs)[1],NewBuyers, replace=TRUE),]
    Traders <- rbind(Traders, Buyers)
    Traders$MarketStatus <- factor(Traders$MarketStatus)
    
    # Buyers search property within their search criteria and calculate their utility
    Buyers <- Traders[Traders$MarketStatus=="Buyer",]      #Buyers is a dataset with all buyers
    LandOnSale <- Parcels[Parcels$OnSale,]

    for(i in 1:dim(Buyers)[1]) {
      BUYER <- Buyers[i,]
      maxBud <- BUYER$Budget
      minBud <- BUYER$minBudget
      
      # ADD OTHER BUYER PREFERENCES HERE
      IDinBud <- LandOnSale$ID[LandOnSale$price<maxBud      
                               & LandOnSale$price*kH*BUYER$Financed/100 + LandOnSale$FloodIns*BUYER$FloodIns  < BUYER$Income * 0.3  # annual amount spent on housing should not be higher than 30% of total income. Control for flood insurance costs if needed
                               & LandOnSale$price>minBud                               # property should be within the budget range of the buyer
                               & (LandOnSale$pflood==0.01)*BUYER$AVOID100FZ==0              # buyer checks if he should avoid a property, given its risk avoidance strategy and given the flood zonation of the property
                               & (LandOnSale$pflood==0.002)*BUYER$AVOID500FZ==0              # buyer checks if he should avoid a property, given its risk avoidance strategy and given the flood zonation of the property
                               ]
      # ADD OTHER BUYER PREFERENCES HERE
      
      NinBud <- length(IDinBud)
      sampleN <- if(NinBud < nN) {NinBud} else {nN}
      PID <- if(sampleN == 1) {IDinBud} else {sample(IDinBud,sampleN)}
      Houses <- Parcels[PID,]
      
      #CALCULATE UTILITY (this function needs updating after discussing with Tatiana)
      #update numbers based on BuyerPrefs data frame
      UThouse <- BUYER$AGEnorm*Houses$AGEnorm +
        BUYER$SQFTnorm*Houses$SQFTnorm +
        BUYER$ACRESnorm*Houses$ACRESnorm +
        BUYER$BEDRMnorm*Houses$BEDRMnorm +
        BUYER$RESIDnorm*Houses$RESIDnorm
      UTILITY <- UThouse
      UTILITY <- round(UTILITY, 2)
      if(NinBud>0) {
        minprice <- min(Houses$price[UTILITY==max(UTILITY)])
        Buyers$PID[i] <- PID[UTILITY==max(UTILITY)&Houses$price==minprice]
        Buyers$Utility[i] <- max(UTILITY)
        Buyers$House.PFLOOD[i] <- Parcels$pflood[Buyers$PID[i]]
        #update perceived likelihood of damage for the property of interest
        rules <- Buyers$House.PFLOOD[i]==LikelyDamage.Buy$House.PFLOOD &
          Buyers$Learning.NewsMedia[i]==LikelyDamage.Buy$Learning.NewsMedia &
          Buyers$Exp.Damage[i]==LikelyDamage.Buy$Exp.Damage2
        if(any(rules)) {Buyers$LikelyDamage[i] <- sample(1:4,1,prob=LikelyDamage.Buy[rules,c("1","2","3","4")])}
        Buyers$Prob.FloodIns[i] <- 0.145+0.07*Buyers$Effect.FloodIns[i]-0.053*Buyers$Costs.FloodIns[i]+0.126*Buyers$LikelyDamage[i]+22.4*Buyers$House.PFLOOD[i]
      } else {
        Buyers$PID[i] <- 0
        Buyers$Utility[i] <- 0
      }
    }
      
    # buyers that cannot find a property within their search criteria leave the market
    Traders[Traders$MarketStatus=="Buyer",] <- Buyers
    Traders <- Traders[!(Traders$MarketStatus=="Buyer" & Traders$PID == 0),]
    Buyers <- Buyers[Buyers$PID!=0,]
    
    # Buyers form bid price
    ReservationPrice <- Buyers$Budget
    askPr <- Parcels$price[Buyers$PID]
    delta <- round(askPr*0.05, -2)
    BidPriceLow <- askPr-delta
    BidPriceHigh <- askPr+delta
    # Maximum bid price cannot exceed the buyer's budget
    BidPriceHigh[BidPriceHigh>ReservationPrice] <- ReservationPrice[BidPriceHigh>ReservationPrice]
    NBuy <- dim(Buyers)[1]
    Buyers$MyPrice <- round(runif(NBuy, BidPriceLow,BidPriceHigh),-3)
    Traders[Traders$MarketStatus=="Buyer",] <- Buyers
    
    # Sellers check their bids and negotiate
    PotBuyList <- unique(Buyers$PID)
    PotBuyList <- sort(PotBuyList)
    Sellers <- Traders[Traders$MarketStatus=="Seller",]
    #no offers - sellers set refused to +1, trade code 7
    Sellers$Refused[!(Sellers$PID %in% PotBuyList)] <- Sellers$Refused[!(Sellers$PID %in% PotBuyList)]+1
    Trades$TradeCode[!(Sellers$PID %in% PotBuyList)] <- 7
    #sellers with offers check the highest bid
    Prices <- Sellers$MyPrice[Sellers$PID %in% PotBuyList]
    maxBid <- rep(NA, length(Prices))
    nBids <- rep(NA, length(Prices))
    for (i in PotBuyList) {
      offerlist <- Buyers$MyPrice[Buyers$PID %in% i]
      max <- max(offerlist)
      maxBid[PotBuyList==i] <- max
      nBids[PotBuyList==i] <- length(offerlist)
    }
    Trades$BidPrice[Sellers$PID %in% PotBuyList] <- maxBid
    Trades$NofBuyers[Sellers$PID %in% PotBuyList] <- nBids
    #bid higher than ask price - trade code 1
    TC1ID <- PotBuyList[maxBid >= Prices]
    TC1Price <- maxBid[maxBid >= Prices]
    if (length(TC1ID)>0) {
      Trades$AskPrice[Trades$ID %in% TC1ID] <- TC1Price
      Trades$TradeCode[Trades$ID %in% TC1ID] <- 1
      Sellers$PID[Sellers$PID %in% TC1ID] <- 0
      for (i in 1:length(TC1ID)) {
        Buyers$MarketStatus[Buyers$PID==TC1ID[i] & Buyers$MyPrice==TC1Price[i]][1] <- "Traded"
      }
      Buyers$PID[Buyers$PID %in% TC1ID & Buyers$MarketStatus == "Buyer"] <- 0
    }
    #bid lower than ask price, nBids>1, buyer raises bid - trade code 2
    TC2ID <- PotBuyList[Prices-maxBid < avRent & maxBid < Prices & nBids > 1]
    TC2Price <- Prices[Prices-maxBid < avRent & maxBid < Prices & nBids > 1]
    TC2PrevBid <- maxBid[Prices-maxBid < avRent & maxBid < Prices & nBids > 1]
    if (length(TC2ID)>0) {
      Trades$BidPrice[Trades$ID %in% TC2ID] <- TC2Price
      Trades$TradeCode[Trades$ID %in% TC2ID] <- 2
      Sellers$PID[Sellers$PID %in% TC2ID] <- 0
      for (i in 1:length(TC2ID)) {
        Buyers$MarketStatus[Buyers$PID==TC2ID[i] & Buyers$MyPrice==TC2PrevBid[i]][1] <- "Traded"
      }
      Buyers$PID[Buyers$PID %in% TC2ID & Buyers$MarketStatus == "Buyer"] <- 0  
    }
    #bid lower than ask price, nBids>1, seller lowers price - trade code 3
    TC3ID <- PotBuyList[Prices-maxBid <= Sellers$aDelta[Trades$TradeCode!=7] & maxBid < Prices & nBids > 1 & Prices-maxBid >= avRent]
    TC3Price <- maxBid[Prices-maxBid <= Sellers$aDelta[Trades$TradeCode!=7] & maxBid < Prices & nBids > 1 & Prices-maxBid >= avRent]
    if (length(TC3ID)>0) {
      Trades$AskPrice[Trades$ID %in% TC3ID] <- TC3Price
      Trades$TradeCode[Trades$ID %in% TC3ID] <- 3
      Sellers$PID[Sellers$PID %in% TC3ID] <- 0
      for (i in 1:length(TC3ID)) {
        Buyers$MarketStatus[Buyers$PID==TC3ID[i] & Buyers$MyPrice==TC3Price[i]][1] <- "Traded"
      }
      Buyers$PID[Buyers$PID %in% TC3ID & Buyers$MarketStatus == "Buyer"] <- 0
    }
    #no trade because traders don't agree, seller has more than 1 bid - trade code 4
    TC4ID <- PotBuyList[Prices-maxBid > avRent & maxBid < Prices & nBids > 1 & Prices-maxBid > Sellers$aDelta[Trades$TradeCode!=7]]
    Trades$TradeCode[Trades$ID %in% TC4ID] <- 4
    Buyers$PID[Buyers$PID %in% TC4ID] <- 0
    #seller has one bid and lowers ask price - trade code 5
    TC5ID <- PotBuyList[Prices-maxBid <= Sellers$aDelta[Trades$TradeCode!=7] & nBids==1 & Prices > maxBid]
    TC5Price <- maxBid[Prices-maxBid <= Sellers$aDelta[Trades$TradeCode!=7] & nBids==1 & Prices > maxBid]
    if (length(TC5ID)>0) {
      Trades$AskPrice[Trades$ID %in% TC5ID] <- TC5Price
      Trades$TradeCode[Trades$ID %in% TC5ID] <- 5
      Sellers$PID[Sellers$PID %in% TC5ID] <- 0
      for (i in 1:length(TC5ID)) {
        Buyers$MarketStatus[Buyers$PID==TC5ID[i] & Buyers$MyPrice==TC5Price[i]] <- "Traded"
      }
    }
    #selles has one bid and traders don't agree - trade code 6
    TC6ID <- PotBuyList[nBids==1 & Prices-maxBid > Sellers$aDelta[Trades$TradeCode!=7]]
    Trades$TradeCode[Trades$ID %in% TC6ID] <- 6
    Buyers$PID[Buyers$PID %in% TC6ID] <- 0
    #set Refused to Refused + 1
    Sellers$Refused[Trades$TradeCode==4 | Trades$TradeCode==6] <- Sellers$Refused[Trades$TradeCode==4 | Trades$TradeCode==6] + 1
    Buyers$Refused[Buyers$PID==0] <- Buyers$Refused[Buyers$PID==0] + 1
    
    #------------------
    # 47.5% of sellers become buyers (Assessed empirically: number of people in the surveys that remain in the same area)
    
    SellersNB <- Sellers[Sellers$PID == 0,]
    NSales <- dim(SellersNB)[1]
    NewBuyerOS <- runif(NSales) < .475
    SellersNB$MarketStatus[NewBuyerOS] <- "Buyer"
    SellersNB <- SellersNB[SellersNB$MarketStatus=="Buyer",]
    NewBuyerOS <- dim(SellersNB)[1]
    SellersNB$MyPrice <- rep(0,NewBuyerOS)
    pAV100 <- 0.04776*SellersNB$FloodIns*SellersNB$Costs.FloodIns+
      0.04464*SellersNB$Effect.MovingOut-0.04560*SellersNB$Costs.MovingOut+
      0.08091*SellersNB$LikelyDamage+0.04438*SellersNB$FEARscore+
      0.16163*SellersNB$Exp.Damage-0.12097
    SellersNB$AVOID100FZ <- runif(NewBuyerOS) < pAV100                                       # a number of sellers avoid the 100-year flood zone in the future
    SellersNB$AVOID500FZ[SellersNB$AVOID100FZ] <- runif(sum(SellersNB$AVOID100FZ)) < 0.565   # a number of sellers also avoid the 500-year flood zone in the future
    SellersNB$yM <- 0
    
    # other sellers leave
    Sellers <- Sellers[Sellers$PID!=0,]
    # reconstruct trader dataset
    NewOwners <- Buyers[Buyers$MarketStatus=="Traded",]
    NewOwners <- NewOwners[order(NewOwners$PID),]
    NewOwners$Debt <- NewOwners$MyPrice*NewOwners$Financed/100
    NewOwners$FloodIns <- runif(nrow(NewOwners)) < NewOwners$Prob.FloodIns
    Buyers <- rbind(Buyers, SellersNB)
    Traders <- rbind(Traders[Traders$MarketStatus=="Traded",],NewOwners,Sellers)
    Traders <- Traders[order(Traders$PID),]
    
    Buyers <- Buyers[Buyers$MarketStatus=="Buyer",]
    Buyers$MyPrice <- 0
    Buyers$Utility <- 0
    
    Traders <- rbind(Traders, Buyers)
    # Traded parcels are not on sale anymore
    Parcels$OnSale[Parcels$OnSale] <- Trades$TradeCode == 6 | Trades$TradeCode == 4 | Trades$TradeCode == 7
    
    # register trades
    NewOwnerBudget <- NewOwners$Budget
    Sales <- data.frame(TICK, Trades$ID, Trades$AskPrice,NA)
    names(Sales) <- names(SALES)
    Sales <- Sales[Trades$AskPrice==Trades$BidPrice,]
    Sales$BuyerBudget <- NewOwnerBudget
    SALES <- rbind(SALES, Sales)
    TRADES <- rbind(TRADES, Trades)
    setwd(tempdir())
    write.table(TRADES, "All trades.csv", sep=",")
    write.table(SALES, "Trade results.csv", sep=",")
    setwd(dir)
    TEMP <- cbind(rep(TICK,NParcels),Traders[Traders$MarketStatus!="Buyer",])
    names(TEMP)[1] <- "TICK"
    # reduce the size of TEMP dataset
    TEMP <- TEMP[c("TICK","PID","House.PFLOOD","Income","Exp.Flood","Exp.Damage","FloodIns","FEARscore","Mortgage","Debt","SellerStatus")]
    TEMP <- TEMP[TEMP$TICK>90,]
    TEMP$Debt <- round(TEMP$Debt,-2)
    TEMP$Income <- round(TEMP$Income,-2)
    HOMEOWNERINFO <- rbind(HOMEOWNERINFO,TEMP)
    remove(TEMP)
  }
  
  #export settings
    write.table(HOMEOWNERINFO,paste0(outputDir,"/Details_",SIM.ID,".csv"), sep=",")
    #drop_upload(file=paste0("Details_",SIM.ID,".csv"),path=outputDir,autorename = FALSE)
    
    WriteResults <- function(SIM.ID) {
      setwd(tempdir())
      Sales <- read.table("Trade results.csv", sep=",", header=TRUE)
      setwd(dir)
      # Create a unique file name
      fileName <- paste("SIM_ID_", SIM.ID,".csv", sep="")
      # Write the data to a temporary file locally
      filePath <- file.path(outputDir, fileName)
      write.csv(Sales, filePath, row.names = FALSE, quote = TRUE)
    }
    
    WriteResults(SIM.ID)
    
  
  
  #debugging
  print(Settings)
  
  print(Sys.time())
}
